## Activipy --- ActivityStreams 2.0 implementation and validator for Python
## Copyright © 2015 Christopher Allan Webber <cwebber@dustycloud.org>
##
## This file is part of Activipy, which is GPLv3+ or Apache v2, your option
## (see COPYING); since that means effectively Apache v2 here's those headers
##
## Apache v2 header:
##   Licensed under the Apache License, Version 2.0 (the "License");
##   you may not use this file except in compliance with the License.
##   You may obtain a copy of the License at
##
##       http://www.apache.org/licenses/LICENSE-2.0
##
##   Unless required by applicable law or agreed to in writing, software
##   distributed under the License is distributed on an "AS IS" BASIS,
##   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##   See the License for the specific language governing permissions and
##   limitations under the License.

import json

import dbm

from activipy import core, vocab


class JsonDBM(object):
    """
    json wrapper around a gdbm database
    """
    def __init__(self, db):
        self.db = db

    def __getitem__(self, key):
        return json.loads(self.db[key.encode('utf-8')].decode('utf-8'))

    def __setitem__(self, key, value):
        self.db[key.encode('utf-8')] = json.dumps(value)

    def __delitem__(self, key):
        del self.db[key.encode('utf-8')]

    def __contains__(self, key):
        return key in self.db

    @classmethod
    def open(cls, filename):
        return cls(dbm.open(filename, 'c'))

    def close(self):
        self.db.close()

    def get(self, key, default=None):
        if key in self.db:
            return self[key]
        else:
            return default

    def fetch_asobj(self, env):
        return core.ASObj(self[id], env)

# Each of these returns the full object inserted into dbm

def dbm_fetch(id, db, env):
    return core.ASObj(db[id], env)

def dbm_save(asobj, db):
    assert asobj.id is not None
    new_val = asobj.json()
    db[asobj.id] = new_val
    return new_val

def dbm_delete(asobj, db):
    assert asobj.id is not None
    del db[asobj.id]


dbm_save_method = core.MethodId(
    "save", "Save object to the DBM store.",
    core.handle_one)
dbm_delete_method = core.MethodId(
    "delete", "Delete object from the DBM store.",
    core.handle_one)

DbmEnv = core.Environment(
    vocabs=[vocab.CoreVocab],
    methods={
        (dbm_save_method, vocab.Object): dbm_save,
        (dbm_delete_method, vocab.Object): dbm_delete},
    shortids=core.shortids_from_vocab(vocab.CoreVocab),
    c_accessors=core.shortids_from_vocab(vocab.CoreVocab))


def dbm_activity_normalized_save(asobj, db):
    assert asobj.id is not None
    as_json = asobj.json()

    def maybe_normalize(key):
        val = as_json.get(key)
        # Skip if not a dictionary with a "@type"
        if not isinstance(val, dict) or not "@type" in val:
            return

        val_asobj = core.ASObj(val, asobj.env)
        # yup, time to normalize
        if asobj.env.is_astype(val_asobj, vocab.Object, inherit=True):
            # If there's no id, then okay, don't normalize
            if val_asobj.id is None:
                return

            if val_asobj.id not in db:
                # save to the database
                asobj.env.asobj_run_method(val_asobj, dbm_save_method, db)

            # and set the key to be the .id
            as_json[key] = val_asobj.id

    maybe_normalize("actor")
    maybe_normalize("object")
    maybe_normalize("target")
    db[asobj.id] = as_json
    return as_json


dbm_denormalize_method = core.MethodId(
    "denormalize", "Expand out an activitystreams object recursively",
    # @@: Should this be a handle_fold?
    core.handle_one)


def dbm_denormalize_object(asobj, db):
    # For now, on any standard object, just return that as-is
    return asobj


def dbm_denormalize_activity(asobj, db):
    as_json = asobj.json()

    def maybe_denormalize(key):
        val = as_json.get(key)
        # If there's no specific val,
        # it's not a string, or it's not in the database,
        # just leave it!
        if val is None or not isinstance(val, str) or val not in db:
            return

        # Otherwise, looks like that value *is* in the database... hey!
        # Let's pull it out and set it as the key.
        as_json[key] = db[val]

    maybe_denormalize("actor")
    maybe_denormalize("object")
    maybe_denormalize("target")
    return core.ASObj(as_json, asobj.env)


DbmNormalizedEnv = core.Environment(
    vocabs=[vocab.CoreVocab],
    methods={
        (dbm_save_method, vocab.Object): dbm_save,
        (dbm_save_method, vocab.Activity): dbm_activity_normalized_save,
        (dbm_delete_method, vocab.Object): dbm_delete,
        (dbm_denormalize_method, vocab.Object): dbm_denormalize_object,
        (dbm_denormalize_method, vocab.Activity): dbm_denormalize_activity},
    shortids=core.shortids_from_vocab(vocab.CoreVocab),
    c_accessors=core.shortids_from_vocab(vocab.CoreVocab))


def dbm_fetch_denormalized(id, db, env):
    """
    Fetch a fully denormalized ASObj from the database.
    """
    return env.asobj_run_method(
        dbm_fetch(id, db, env),
        dbm_denormalize_method, db)



## Activipy --- ActivityStreams 2.0 implementation and validator for Python
## Copyright © 2015 Christopher Allan Webber <cwebber@dustycloud.org>
##
## This file is part of Activipy, which is GPLv3+ or Apache v2, your option
## (see COPYING); since that means effectively Apache v2 here's those headers
##
## Apache v2 header:
##   Licensed under the Apache License, Version 2.0 (the "License");
##   you may not use this file except in compliance with the License.
##   You may obtain a copy of the License at
##
##       http://www.apache.org/licenses/LICENSE-2.0
##
##   Unless required by applicable law or agreed to in writing, software
##   distributed under the License is distributed on an "AS IS" BASIS,
##   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##   See the License for the specific language governing permissions and
##   limitations under the License.



## Activipy --- ActivityStreams 2.0 implementation and validator for Python
## Copyright © 2015 Christopher Allan Webber <cwebber@dustycloud.org>
##
## This file is part of Activipy, which is GPLv3+ or Apache v2, your option
## (see COPYING); since that means effectively Apache v2 here's those headers
##
## Apache v2 header:
##   Licensed under the Apache License, Version 2.0 (the "License");
##   you may not use this file except in compliance with the License.
##   You may obtain a copy of the License at
##
##       http://www.apache.org/licenses/LICENSE-2.0
##
##   Unless required by applicable law or agreed to in writing, software
##   distributed under the License is distributed on an "AS IS" BASIS,
##   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##   See the License for the specific language governing permissions and
##   limitations under the License.

from activipy.core import (
    ASType, ASVocab, Environment, shortids_from_vocab,
    chain_dicts, make_simple_loader)
from activipy import vocab

def checkup_uri(identifier):
    return "http://checkup.example/ns#" + identifier

CheckIn = ASType(
    checkup_uri("CheckIn"), [vocab.Arrive], "CheckIn",
    notes=(
        "Check in to a location."))

Coupon = ASType(
    checkup_uri("Coupon"), [vocab.Object], "Coupon",
    notes=(
        "A redeemable voucher (by redeem_url) for hopefully "
        "something exciting!"))

RoyalStatus = ASType(
    checkup_uri("RoyalStatus"), [vocab.Object], "RoyalStatus",
    notes=(
        "How royal you are at a given location!"))


CheckUpVocab = ASVocab([CheckIn, Coupon, RoyalStatus])

CHECKUP_EXTRA_CONTEXT_NAMESPACED = {"CheckUp": "http://checkup.example/ns#"}

CheckUpNSEnv = Environment(
    vocabs=[vocab.CoreVocab],
    shortids=chain_dicts(
        shortids_from_vocab(vocab.CoreVocab),
        shortids_from_vocab(CheckUpVocab, "CheckUp")),
    c_accessors=chain_dicts(
        shortids_from_vocab(vocab.CoreVocab),
        shortids_from_vocab(CheckUpVocab)),
    extra_context=CHECKUP_EXTRA_CONTEXT_NAMESPACED)

CHECKUP_EXTRA_CONTEXT_VERBOSE = {
    "CheckIn": {
        "@id": CheckIn.id_uri,
        "@type": "@id"},
    "Coupon": {
        "@id": Coupon.id_uri,
        "@type": "@id"},
    "RoyalStatus": {
        "@id": RoyalStatus.id_uri,
        "@type": "@id"}}

CheckUpVerboseEnv = Environment(
    vocabs=[vocab.CoreVocab],
    shortids=chain_dicts(
        shortids_from_vocab(vocab.CoreVocab),
        shortids_from_vocab(CheckUpVocab)),
    c_accessors=chain_dicts(
        shortids_from_vocab(vocab.CoreVocab),
        shortids_from_vocab(CheckUpVocab)),
    extra_context=CHECKUP_EXTRA_CONTEXT_VERBOSE)


CHECKUP_EXTRA_CONTEXT_URI = "http://checkup.example/context.jld"

CHECKUP_JSONLD_LOADER = make_simple_loader(
    {CHECKUP_EXTRA_CONTEXT_URI: {"@context": CHECKUP_EXTRA_CONTEXT_VERBOSE}},
    # @@: Do we want this in the demo?
    load_unknown_urls=False)

CheckUpEnv = Environment(
    vocabs=[vocab.CoreVocab],
    shortids=chain_dicts(
        shortids_from_vocab(vocab.CoreVocab),
        shortids_from_vocab(CheckUpVocab)),
    c_accessors=chain_dicts(
        shortids_from_vocab(vocab.CoreVocab),
        shortids_from_vocab(CheckUpVocab)),
    extra_context=CHECKUP_EXTRA_CONTEXT_URI,
    document_loader=CHECKUP_JSONLD_LOADER)



## Activipy --- ActivityStreams 2.0 implementation and validator for Python
## Copyright © 2015 Christopher Allan Webber <cwebber@dustycloud.org>
##
## This file is part of Activipy, which is GPLv3+ or Apache v2, your option
## (see COPYING); since that means effectively Apache v2 here's those headers
##
## Apache v2 header:
##   Licensed under the Apache License, Version 2.0 (the "License");
##   you may not use this file except in compliance with the License.
##   You may obtain a copy of the License at
##
##       http://www.apache.org/licenses/LICENSE-2.0
##
##   Unless required by applicable law or agreed to in writing, software
##   distributed under the License is distributed on an "AS IS" BASIS,
##   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##   See the License for the specific language governing permissions and
##   limitations under the License.

from pkg_resources import resource_filename
import copy
import json

from pyld import jsonld


# The actual instances of these are defined in vocab.py

class ASType(object):
    """
    A @type than an ActivityStreams object might take on.

    BTW, you might wonder why this isn't using python class heirarchies
    as an abstraction.  The reason is simple: ActivityStreams objects
    can have multiple types listed under @type.  So our inheritance
    model is a bit different than python's.
    """
    def __init__(self, id_uri, parents, id_short=None, notes=None):
        self.id_uri = id_uri
        self.parents = parents
        self.id_short = id_short
        self.notes = notes

        self._inheritance = None

    def validate(self, asobj):
        validator = self.methods.get("validate")
        if validator is not None:
            validator(asobj)

    def __repr__(self):
        return "<ASType %s>" % (self.id_short or self.id_uri)

    # TODO: Use a generic memoizer?
    @property
    def inheritance_chain(self):
        # memoization
        if self._inheritance is None:
            self._inheritance = astype_inheritance_list(self)

        return self._inheritance

    def __call__(self, id=None, env=None, **kwargs):
        # @@: Is this okay?  Kinda bold!
        if env is None:
            from activipy import vocab
            env = vocab.BasicEnv

        if self in env.shortids_reversemap:
            type_val = env.shortids_reversemap[self]
        else:
            type_val = self.id_uri

        jsobj = {"@type": type_val}
        jsobj.update(kwargs)
        if id:
            jsobj["@id"] = id

        if env:
            return ASObj(jsobj, env=env)
        else:
            return ASObj(jsobj)


def astype_inheritance_list(*astypes):
    """
    Gather the inheritance list for an ASType or multiple ASTypes

    We need this because unlike w/ Python classes, an individual
    ASObj can have composite types.
    """
    def traverse(astype, family):
        family.append(astype)
        for parent in astype.parents:
            traverse(parent, family)

        return family

    # not deduped at this point
    family = []
    for astype in astypes:
        family = traverse(astype, family)

    # okay, dedupe here, only keep the oldest instance of each
    family.reverse()
    deduped_family = []
    for member in family:
        if member not in deduped_family:
            deduped_family.append(member)

    deduped_family.reverse()
    return deduped_family


class ASVocab(object):
    """
    Mapping of known type IDs to ASTypes

    TODO: Maybe this should include the appropriate context
      it's working within?
    """
    def __init__(self, vocabs):
        self.vocab_map = self._map_vocabs(vocabs)

    def _map_vocabs(self, vocabs):
        return {
            type.id_uri: type
            for type in vocabs}


# TODO: Add this one by default
AS2_CONTEXT_FILE = resource_filename(
    'activipy', 'activitystreams2-context.jsonld')
AS2_CONTEXT = json.loads(open(AS2_CONTEXT_FILE, 'r').read())
AS2_CONTEXT_URI = (
    "http://www.w3.org/TR/activitystreams-core/activitystreams2-context.jsonld")
AS2_DEFAULT_URL_MAP = {
    AS2_CONTEXT_URI: AS2_CONTEXT}

# Once things are cached, json-ld expansion seems to happen at about
# 1250 douments / second on my laptop

def make_simple_loader(url_map, load_unknown_urls=True,
                       cache_externally_loaded=True):
    def _make_context(url, doc):
        return {
            "contextUrl": None,
            "documentUrl": url,
            "document": doc}

    # Wrap in the structure that's expected to come back from the
    # documentLoader
    _pre_url_map = {}
    _pre_url_map.update(AS2_DEFAULT_URL_MAP)
    _pre_url_map.update(url_map)
    _url_map = {
        url: _make_context(url, doc)
        for url, doc in _pre_url_map.items()}

    def loader(url):
        if url in _url_map:
            return _url_map[url]
        elif load_unknown_urls:
            doc = jsonld.load_document(url)
            # @@: Is this optimization safe in all cases?
            if isinstance(doc["document"], str):
                doc["document"] = json.loads(doc["document"])
            if cache_externally_loaded:
                _url_map[url] = doc
            return doc
        else:
            raise jsonld.JsonLdError(
                "url not found and loader set to not load unknown URLs.",
                {'url': url})

    return loader

default_loader = make_simple_loader({})

# TODO: This was a good early in-comments braindump; now move to the
# documentation and restructure!

# So, questions for ourselves.  What is this, if not merely a json
# object?  After all, an ActivityStreams object can be represented as
# "just JSON", and be done with it.  So what's *useful*?
#
# Here are some potentially useful properties:
#  - Expanded json-ld form
#  - Extracted types
#    - As short forms
#    - As expanded / unambiguous URIs (see json-ld)
#    - As ASType objects (where possible)
#  - Validation
#  - Lookup of what a property key "means"
#    (checking against activitystreams vocabulary)
#  - key-value access, including fetching any nested activitystreams
#    objects as ASObj types
#  - json serialization to string
#
# Of all the above, it would be nice not to have to repeat these
# operations.  If we've done it once, that should be good enough
# forever... in other words, memoization.  But memoization means
# that the object should be immutable.
#
# ... but maybe ASObj objects *should* be immutable.
# This means we copy.deepcopy() on our way in, and if users want
# to change things, they either make a new ASObj or get back
# entirely new ASObj objects.
#
# I like this idea...

class ASObj(object):
    """
    The general ActivityStreams object that a user will work with
    """
    def __init__(self, jsobj, env=None):
        if not env:
            from activipy import vocab
            env = vocab.BasicEnv
        self.env = env

        self.__jsobj = deepcopy_jsobj_in(jsobj, env)
        assert (isinstance(self.__jsobj.get("@type"), str) or
                isinstance(self.__jsobj.get("@type"), list))

        self.m = self.env._build_m_map(self)

    def __getitem__(self, key):
        val = self.__jsobj[key]
        if isinstance(val, dict) and "@type" in val:
            return ASObj(val, self.env)
        else:
            return deepcopy_jsobj_out(val, env=self.env)

    # META TODO: Convert some @property here to @memoized_property
    @property
    def types(self):
        type_attr = self["@type"]
        if isinstance(self["@type"], list):
            return type_attr
        else:
            return [type_attr]

    @property
    def types_expanded(self):
        return copy.deepcopy(self.__expanded()[0]["@type"])

    # TODO: Memoize
    @property
    def types_astype(self):
        return self.env.asobj_astypes(self)

    # TODO: Memoize
    @property
    def types_inheritance(self):
        return self.env.asobj_astype_inheritance(self)

    # Don't memoize this, users might mutate
    def json(self):
        return copy.deepcopy(self.__jsobj)

    # TODO: Memoize
    def json_str(self):
        return json.dumps(self.json())

    # TODO Memoize
    def __expanded(self):
        if self.env.document_loader:
            document_loader = self.env.document_loader
        else:
            document_loader = default_loader

        options = {"expandContext": self.env.implied_context}
        if document_loader:
            options["documentLoader"] = document_loader

        return jsonld.expand(self.__jsobj, options)

    def expanded(self):
        """
        Note: this produces a copy of the object returned, so consumers
          of this method may want to keep a copy of its result
          rather than calling over and over.
        """
        return copy.deepcopy(self.__expanded())

    # TODO: Memoize
    def expanded_str(self):
        return json.dumps(self.expanded())

    @property
    def id(self):
        return self.__jsobj.get("@id")

    def __repr__(self):
        if self.id:
            return "<ASObj %s \"%s\">" % (
                ", ".join(self.types),
                self.id)
        else:
            return "<ASObj %s>" % ", ".join(self.types)


def deepcopy_jsobj_base(jsobj, env, going_in=True):
    """
    Perform a deep copy of a JSON style object
    """
    going_out = not going_in

    def add_context(this_dict):
        if env.extra_context is not None:
            this_dict["@context"] = env.extra_context

    def remove_context(this_dict):
        if "@context" in this_dict:
            del this_dict["@context"]
        return this_dict

    def copy_asobj(asobj):
        if going_in:
            return remove_context(asobj.json())
        else:
            return asobj

    def copy_dict(this_dict):
        # Looks like an ASObj
        if going_out and "@type" in this_dict:
            return ASObj(this_dict, env)

        # Otherwise, just recursively copy the dict
        new_dict = {}
        for key, val in this_dict.items():
            new_dict[key] = copy_main(val)
        return new_dict

    def copy_list(this_list):
        new_list = []
        for item in this_list:
            new_list.append(copy_main(item))
        return new_list

    def copy_main(jsobj):
        if isinstance(jsobj, dict):
            return copy_dict(jsobj)
        elif isinstance(jsobj, ASObj):
            return copy_asobj(jsobj)
        elif isinstance(jsobj, list):
            return copy_list(jsobj)
        else:
            # All other JSON type objects are immutable,
            # just copy them down.
            # @@: We could provide validation that it's
            #   a valid json object here but that seems like
            #   it would bring unnecessary performance penalties.
            return jsobj

    if going_in:
        # Should be a dictionary or ASObj on the way in for this
        assert isinstance(jsobj, dict) or isinstance(jsobj, ASObj)

    final_json = copy_main(jsobj)

    if going_in:
        add_context(final_json)

    return final_json


def deepcopy_jsobj_in(jsobj, env):
    return deepcopy_jsobj_base(jsobj, env, going_in=True)

def deepcopy_jsobj_out(jsobj, env):
    return deepcopy_jsobj_base(jsobj, env, going_in=False)


# @@: Maybe rename to MethodSpec?
class MethodId(object):
    # TODO: fill in
    """
    A method identifier
    """
    def __init__(self, name, description, handler):
        self.name = name
        self.description = description
        self.handler = handler

    def __repr__(self):
        return "<MethodId %s>" % self.name


class NoMethodFound(Exception): pass

def throw_no_method_error(asobj):
    raise NoMethodFound("Could not find a method for type: %s" % (
        ", ".join(asobj.types)))

def handle_one(astype_methods, asobj, _fallback=throw_no_method_error):
    if len(astype_methods) == 0:
        _fallback(asobj)

    def func(*args, **kwargs):
        method, astype = astype_methods[0]
        return method(asobj, *args, **kwargs)
    return func


def handle_map(astype_methods, asobj):
    def func(*args, **kwargs):
        return [method(asobj, *args, **kwargs)
                for method, astype in astype_methods]
    return func


class HaltIteration(object):
    def __init__(self, val):
        self.val = val


def handle_fold(astype_methods, asobj):
    def func(initial=None, *args, **kwargs):
        val = initial
        for method, astype in astype_methods:
            # @@: Not sure if asobj or val coming first is a better interface...
            val = method(asobj, val, *args, **kwargs)
            # Provide a way to break out of the loop early...?
            # @@: Is this a good idea, or even useful for anything?
            if isinstance(val, HaltIteration):
                val = val.val
                break
        return val
    return func


# TODO
# @@: Can this be just an @property on Environment?
class AttrMapper(object):
    def __init__(self, attrib_map):
        for key, val in attrib_map.items():
            setattr(self, key, val)

class TypeConstructor(object):
    def __init__(self, astype, env):
        self.astype = astype
        self.__env = env

    def __call__(self, *args, **kwargs):
        return self.astype(env=self.__env, *args, **kwargs)

    def __repr__(self):
        return "<TypeConstructor for %s>" % self.astype.__repr__()


class EnvironmentMismatch(Exception):
    """
    Raised when an ASObj calls a method through an Environment
    but does not have that environment bound to itself.
    """
    pass


class Environment(object):
    """
    An environment to collect vocabularies and provide
    methods for activitystream types
    """
    implied_context = AS2_CONTEXT_URI

    def __init__(self, vocabs=None, methods=None,
                 # not ideal, I'd rather somehow load something
                 # that uses the vocabs as passed in, but that
                 # introduces its own complexities
                 shortids=None, c_accessors=None,
                 extra_context=None,
                 document_loader=default_loader):
        self.vocabs = vocabs or []
        self.methods = methods or {}
        # @@: Should we make all short ids mandatorily contain
        #   the base schema?
        self.shortids = shortids or {}
        self.shortids_reversemap = {
            val: key for key, val in self.shortids.items()}
        self.extra_context = extra_context
        self.document_loader = document_loader
        self.c = self.__build_c_accessors(c_accessors or {})
        self.m = self._build_m_map()

        self.uri_map = self.__build_uri_map()

    def __build_c_accessors(self, c_accessors):
        return AttrMapper(
            {name: TypeConstructor(astype, self)
             for name, astype in c_accessors.items()})

    def _build_m_map(self, asobj=None):
        def make_method_dispatcher(method_id):
            def method_dispatcher(asobj, *args, **kwargs):
                method = self.asobj_get_method(asobj, method_id)
                return method(*args, **kwargs)
            if asobj is None:
                return method_dispatcher
            else:
                # in this variation, we already know what the
                # asobj is
                def curried_method_dispatcher(*args, **kwargs):
                    return method_dispatcher(asobj, *args, **kwargs)
                return curried_method_dispatcher

        method_ids = set([method_id for (method_id, astype)
                          in self.methods.keys()])
        m_mapping = {
            method_id.name: make_method_dispatcher(method_id)
            for method_id in method_ids}

        return AttrMapper(m_mapping)

    def __build_uri_map(self):
        uri_map = {}
        for vocab in self.vocabs:
            uri_map.update(vocab.vocab_map)

        return uri_map

    def _process_type_simple(self, type_id):
        # Try by short ID (in short IDs marked as acceptable for this)
        if type_id in self.shortids:
            return self.shortids[type_id]

        # Try by URI
        elif type_id in self.uri_map:
            return self.uri_map[type_id]

        else:
            # this would happen anyway, but might as well be explicit
            # about what's happening here in the code flow
            return None

    def asobj_astypes(self, asobj):
        final_types = []
        process_as_jsonld = False
        for type_id in asobj.types:
            processed_type = self._process_type_simple(type_id)
            if processed_type is not None:
                final_types.append(processed_type)
            else:
                # We have to bail out
                process_as_jsonld = True
                break

        # Are there any remaining types to process here?
        if process_as_jsonld:
            # @@: We could do a version of this which didn't
            #   throw away the information we already had,
            #   maybe.  But it would be tricky.
            final_types = []
            asobj_jsonld = asobj.expanded()
            for type_uri in asobj_jsonld[0]["@type"]:
                processed_type = self._process_type_simple(type_uri)
                if processed_type is not None:
                    final_types.append(processed_type)

        return final_types

    def asobj_astype_inheritance(self, asobj):
        return astype_inheritance_list(
            *self.asobj_astypes(asobj))

    def is_astype(self, asobj, astype, inherit=True):
        """
        Check to see if an ASObj is of ASType; check full inheritance chain
        """
        if not isinstance(asobj, ASObj):
            return False

        if inherit:
            return astype in self.asobj_astype_inheritance(asobj)
        else:
            return astype in self.asobj_astypes(asobj)

    # @@: Should we drop the asobj_ from these method names?
    def asobj_get_method(self, asobj, method):
        if asobj.env is not self:
            raise EnvironmentMismatch(
                "ASObj attempted to call method with an Environment "
                "it was not bound to!")

        # get all types for this asobj
        astypes = self.asobj_astype_inheritance(asobj)

        # get a map of all relevant {method_proc: astype}
        return method.handler(
            [(self.methods[(method, astype)], astype)
             for astype in astypes
             if (method, astype) in self.methods],
            asobj)

    def asobj_run_method(self, asobj, method, *args, **kwargs):
        # make note of why arguments make this slightly lossy
        # when passing on; eg, can't use asobj/method in the
        # arguments to this function
        return self.asobj_get_method(asobj, method)(*args, **kwargs)


def shortids_from_vocab(vocab, prefix=None):
    """
    Get a mapping of all short ids to their ASType objects in a vocab

    Useful for mapping shortids to ASType objects!
    """
    def maybe_add_prefix(id_short):
        if prefix:
            return "%s:%s" % (prefix, id_short)
        else:
            return id_short

    return {
        maybe_add_prefix(v.id_short): v
        for v in vocab.vocab_map.values()}


def chain_dicts(*dicts):
    """
    Chain together a series of dictionaries into one
    """
    final_dict = {}
    for this_dict in dicts:
        final_dict.update(this_dict)
    return final_dict



## Activipy --- ActivityStreams 2.0 implementation and testing for Python
## Copyright © 2015 Christopher Allan Webber <cwebber@dustycloud.org>
##
## This file is part of Activipy, which is GPLv3+ or Apache v2, your option
## (see COPYING); since that means effectively Apache v2 here's those headers
##
## Apache v2 header:
##   Licensed under the Apache License, Version 2.0 (the "License");
##   you may not use this file except in compliance with the License.
##   You may obtain a copy of the License at
##
##       http://www.apache.org/licenses/LICENSE-2.0
##
##   Unless required by applicable law or agreed to in writing, software
##   distributed under the License is distributed on an "AS IS" BASIS,
##   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##   See the License for the specific language governing permissions and
##   limitations under the License.
##
## In addition, this page copies huge swaths of documentation from the
## ActivityStreams 2.0 Vocabulary document,
##   http://www.w3.org/TR/activitystreams-vocabulary/
##
## Copyright © 2015 Activity Streams Working Group, IBM & W3C® (MIT,
##   ERCIM, Keio, Beihang). W3C liability, trademark and permissive
##   document license rules apply. 
##
## which is released under the
## "W3C Software and Document Notice and License":
## 
##    This work is being provided by the copyright holders under the
##    following license.
##
##    License
##    -------
##    
##    By obtaining and/or copying this work, you (the licensee) agree
##    that you have read, understood, and will comply with the
##    following terms and conditions.
##    
##    Permission to copy, modify, and distribute this work, with or
##    without modification, for any purpose and without fee or royalty
##    is hereby granted, provided that you include the following on
##    ALL copies of the work or portions thereof, including
##    modifications:
##    
##     - The full text of this NOTICE in a location viewable to users
##       of the redistributed or derivative work.
##     - Any pre-existing intellectual property disclaimers, notices,
##       or terms and conditions. If none exist, the W3C Software and
##       Document Short Notice should be included.
##     - Notice of any changes or modifications, through a copyright
##       statement on the new code or document such as "This software
##       or document includes material copied from or derived from
##       [title and URI of the W3C document]. Copyright © [YEAR] W3C®
##       (MIT, ERCIM, Keio, Beihang)." 
##    
##    Disclaimers
##    -----------
##    
##    THIS WORK IS PROVIDED "AS IS," AND COPYRIGHT HOLDERS MAKE NO
##    REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED, INCLUDING BUT
##    NOT LIMITED TO, WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY
##    PARTICULAR PURPOSE OR THAT THE USE OF THE SOFTWARE OR DOCUMENT
##    WILL NOT INFRINGE ANY THIRD PARTY PATENTS, COPYRIGHTS,
##    TRADEMARKS OR OTHER RIGHTS.
##    
##    COPYRIGHT HOLDERS WILL NOT BE LIABLE FOR ANY DIRECT, INDIRECT,
##    SPECIAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF ANY USE OF THE
##    SOFTWARE OR DOCUMENT.
##    
##    The name and trademarks of copyright holders may NOT be used in
##    advertising or publicity pertaining to the work without
##    specific, written prior permission. Title to copyright in this
##    work will at all times remain with copyright holders.

from .core import ASType
from .core import ASVocab, Environment, shortids_from_vocab

def as_uri(identifier):
    return "http://www.w3.org/ns/activitystreams#" + identifier


# Core classes
# ============

Object = ASType(
    as_uri("Object"), [], "Object",
    notes=(
        "Describes an object of any kind. "
        "The Object class serves as the base class for most of the "
        "other kinds of objects defined in the Activity Vocabulary, "
        "include other Core classes such as Activity, "
        "IntransitiveActivity, Actor, Collection and OrderedCollection."))

Link = ASType(
    as_uri("Link"), [], "Link",
    notes=(
        "A Link is an indirect, qualified reference to a resource identified by"
        "a URL. The fundamental model for links is established by [RFC5988]. "
        "Many of the properties defined by the Activity Vocabulary allow "
        "values that are either instances of Object or Link. When a Link is "
        "used, it establishes a qualified relation connecting the subject "
        "(the containing object) to the resource identified by the href."))

Activity = ASType(
    as_uri("Activity"), [Object], "Activity",
    notes=(
        "An Activity is a subclass of Object that describes some form of "
        "action that may happen, is currently happening, or has already "
        "happened. The Activity class itself serves as an abstract base "
        "class for all types of activities. It is important to note that "
        "the Activity class itself does not carry any specific semantics "
        "about the kind of action being taken."))

IntransitiveActivity = ASType(
    as_uri("IntransitiveActivity"), [Activity], "IntransitiveActivity",
    notes=(
        "Instances of IntransitiveActivity are a subclass of Activity whose "
        "actor property identifies the direct object of the action as opposed "
        "to using the object property."))

Actor = ASType(
    as_uri("Actor"), [Object], "Actor",
    notes=(
        "An Actor is any entity that is capable of being the primary actor "
        "for an Activity."))

Collection = ASType(
    as_uri("Collection"), [Object], "Collection",
    notes=(
        "A Collection is a subclass of Object that represents ordered or "
        "unordered sets of Object or Link instances.\n\n"
        "Refer to the Activity Streams 2.0 Core specification for a complete"
        "description of the Collection type."))

OrderedCollection = ASType(
    as_uri("OrderedCollection"), [Collection], "OrderedCollection",
    notes=(
        "A subclass of Collection in which members of the logical collection "
        "are assumed to always be strictly ordered."))

CollectionPage = ASType(
    as_uri("CollectionPage"), [Collection], "CollectionPage",
    notes=(
        "Used to represent distinct subsets of items from a Collection. "
        "Refer to the Activity Streams 2.0 Core for a complete description of "
        "the CollectionPage object."))

OrderedCollectionPage = ASType(
    as_uri("OrderedCollectionPage"), [OrderedCollection, CollectionPage],
    "OrderedCollectionPage",
    notes=(
        "Used to represent ordered subsets of items from an OrderedCollection. "
        "Refer to the Activity Streams 2.0 Core for a complete description of "
        "the OrderedCollectionPage object."))



# Extended Classes: Activity Types
# ================================

Accept = ASType(
    as_uri("Accept"), [Activity],
    "Accept",
    notes=(
        "Indicates that the actor accepts the object. "
        "The target property can be used in certain circumstances to indicate "
        "the context into which the object has been accepted. For instance, "
        "when expressing the activity, \"Sally accepted Joe into the Club\", "
        "the \"target\" would identify the \"Club\"."))

TentativeAccept = ASType(
    as_uri("TentativeAccept"), [Accept],
    "TentativeAccept",
    notes=(
        "A specialization of Accept indicating that the acceptance is "
        "tentative."))

Add = ASType(
    as_uri("Add"), [Activity],
    "Add",
    notes=(
        "Indicates that the actor has added the object to the target. If the "
        "target property is not explicitly specified, the target would need "
        "to be determined implicitly by context. The origin can be used to "
        "identify the context from which the object originated."))

Arrive = ASType(
    as_uri("Arrive"), [IntransitiveActivity],
    "Arrive",
    notes=(
        "An IntransitiveActivity that indicates that the actor has arrived "
        "at the location. The origin can be used to identify the context "
        "from which the actor originated. The target typically has no defined "
        "meaning."))

Create = ASType(
    as_uri("Create"), [Activity],
    "Create",
    notes=(
        "Indicates that the actor has created the object."))

Delete = ASType(
    as_uri("Delete"), [Activity],
    "Delete",
    notes=(
        "Indicates that the actor has deleted the object. If specified, "
        "the origin indicates the context from which the object was "
        "deleted."))

Follow = ASType(
    as_uri("Follow"), [Activity],
    "Follow",
    notes=(
        "Indicates that the actor is \"following\" the object. Following is "
        "defined in the sense typically used within Social systems in which "
        "the actor is interested in any activity performed by or on the "
        "object. The target and origin typically have no defined meaning."))

Ignore = ASType(
    as_uri("Ignore"), [Activity],
    "Ignore",
    notes=(
        "Indicates that the actor is ignoring the object. "
        "The target and origin typically have no defined meaning."))

Join = ASType(
    as_uri("Join"), [Activity],
    "Join",
    notes=(
        "Indicates that the actor has joined the object. The target and "
        "origin typically have no defined meaning."))

Leave = ASType(
    as_uri("Leave"), [Activity],
    "Leave",
    notes=(
        "Indicates that the actor has left the object. The target and origin "
        "typically have no meaning."))

Like = ASType(
    as_uri("Like"), [Activity],
    "Like",
    notes=(
        "Indicates that the actor likes, recommends or endorses the object. "
        "The target and origin typically have no defined meaning."))

Offer = ASType(
    as_uri("Offer"), [Activity],
    "Offer",
    notes=(
        "Indicates that the actor is offering the object. If specified, the "
        "target indicates the entity to which the object is being offered."))

Invite = ASType(
    as_uri("Invite"), [Offer],
    "Invite",
    notes=(
        "A specialization of Offer in which the actor is extending an "
        "invitation for the object to the target."))

Reject = ASType(
    as_uri("Reject"), [Activity],
    "Reject",
    notes=(
        "Indicates that the actor is rejecting the object. The target and "
        "origin typically have no defined meaning."))

TentativeReject = ASType(
    as_uri("TentativeReject"), [Reject],
    "TentativeReject",
    notes=(
        "A specialization of Reject in which the rejection is considered "
        "tentative."))

Remove = ASType(
    as_uri("Remove"), [Activity],
    "Remove",
    notes=(
        "Indicates that the actor is removing the object. If specified, the "
        "origin indicates the context from which the object is being removed."))

Undo = ASType(
    as_uri("Undo"), [Activity],
    "Undo",
    notes=(
        "Indicates that the actor is undoing the object. In most cases, "
        "the object will be an Activity describing some previously performed "
        "action (for instance, a person may have previously \"liked\" "
        "an article but, for whatever reason, might choose to undo that "
        "like at some later point in time).\n\n"
        "The target and origin typically have no defined meaning."))

Update = ASType(
    as_uri("Update"), [Activity],
    "Update",
    notes=(
        "Indicates that the actor has updated the object. Note, however, that "
        "this vocabulary does not define a mechanism for describing the "
        "actual set of modifications made to object.\n\n"
        "The target and origin typically have no defined meaning."))

Experience = ASType(
    as_uri("Experience"), [Activity],
    "Experience",
    notes=(
        "Indicates that the actor has experienced the object. The type of "
        "experience is not specified."))

View = ASType(
    as_uri("View"), [Experience],
    "View",
    notes=(
        "Indicates that the actor has viewed the object. Viewing is a "
        "specialization of Experience."))

Listen = ASType(
    as_uri("Listen"), [Experience],
    "Listen",
    notes=(
        "Indicates that the actor has listened to the object. Listening is a "
        "specialization of Experience."))

Read = ASType(
    as_uri("Read"), [Experience],
    "Read",
    notes=(
        "Indicates that the actor has read the object. Reading is a "
        "specialization of Experience."))

Move = ASType(
    as_uri("Move"), [Activity],
    "Move",
    notes=(
        "Indicates that the actor has moved object from origin to target. If "
        "the origin or target are not specified, either can be determined by "
        "context."))

Travel = ASType(
    as_uri("Travel"), [IntransitiveActivity],
    "Travel",
    notes=(
        "Indicates that the actor is traveling to target from origin. "
        "Travel is an IntransitiveObject whose actor specifies the direct "
        "object. If the target or origin are not specified, either can be "
        "determined by context."))

Announce = ASType(
    as_uri("Announce"), [Activity],
    "Announce",
    notes=(
        "Indicates that the actor is calling the target's attention the object."
        "\n\n"
        "The origin typically has no defined meaning."))

Block = ASType(
    as_uri("Block"), [Ignore],
    "Block",
    notes=(
        "Indicates that the actor is blocking the object. Blocking is a "
        "stronger form of Ignore. The typical use is to support social systems "
        "that allow one user to block activities or content of other users. "
        "The target and origin typically have no defined meaning."))

Flag = ASType(
    as_uri("Flag"), [Activity],
    "Flag",
    notes=(
        "Indicates that the actor is \"flagging\" the object. Flagging is "
        "defined in the sense common to many social platforms as reporting "
        "content as being inappropriate for any number of reasons."))

Dislike = ASType(
    as_uri("Dislike"), [Activity],
    "Dislike",
    notes=(
        "Indicates that the actor dislikes the object."))



# Extended Classes: Actor types
# =============================

Application = ASType(
    as_uri("Application"), [Actor],
    "Application",
    notes=(
        "Describes a software application."))

Group = ASType(
    as_uri("Group"), [Actor],
    "Group",
    notes=(
        "Represents a formal or informal collective of Actors."))

Organization = ASType(
    as_uri("Organization"), [Actor],
    "Organization",
    notes=(
        "Represents an organization."))

Person = ASType(
    as_uri("Person"), [Actor],
    "Person",
    notes=(
        "Represents an individual person."))

Process = ASType(
    as_uri("Process"), [Actor],
    "Process",
    notes=(
        "Represents a series of actions taken to achieve a particular goal."))

Service = ASType(
    as_uri("Service"), [Actor],
    "Service",
    notes=(
        "Represents a service of any kind."))



# Extended Classes: Object Types
# ==============================

Relationship = ASType(
    as_uri("Relationship"), [Object],
    "Relationship",
    notes=(
        "Describes a relationship between two individuals. "
        "The subject and object properties are used to identify the "
        "connected individuals.\n\n"
        "See 3.3.1 [of ActivityStreams 2.0 Vocabulary document] Representing "
        "Relationships Between Entities for additional information."))

Content = ASType(
    as_uri("Content"), [Object],
    "Content",
    notes=(
        "Describes an entity representing any form of content. Examples "
        "include documents, images, etc. Content objects typically are not "
        "able to perform activities on their own, yet rather are usually the "
        "object or target of activities."))

Article = ASType(
    as_uri("Article"), [Content],
    "Article",
    notes=(
        "Represents any kind of multi-paragraph written work."))

Album = ASType(
    as_uri("Album"), [Collection],
    "Album",
    notes=(
        "A type of Collection typically used to organize Image, Video or Audio "
        "objects."))

Folder = ASType(
    as_uri("Folder"), [Collection],
    "Folder",
    notes=(
        "A type of Collection typically used to organize objects such as"
        "Documents."))

Story = ASType(
    as_uri("Story"), [OrderedCollection],
    "Story",
    notes=(
        "A type of Ordered Collection usually containing Content Items "
        "organized to \"tell a story\". "))

Document = ASType(
    as_uri("Document"), [Content],
    "Document",
    notes=(
        "Represents a document of any kind."))

Audio = ASType(
    as_uri("Audio"), [Document],
    "Audio",
    notes=(
        "Represents an audio document of any kind."))

Image = ASType(
    as_uri("Image"), [Document],
    "Image",
    notes=(
        "An image document of any kind"))

Video = ASType(
    as_uri("Video"), [Content],
    "Video",
    notes=("Represents a video document of any kind."))

Note = ASType(
    as_uri("Note"), [Content],
    "Note",
    notes=(
        "Represents a short work typically less than a single "
        "paragraph in length."))

Page = ASType(
    as_uri("Page"), [Document],
    "Page",
    notes=(
        "Represents a Web Page."))

Question = ASType(
    as_uri("Question"), [Content, IntransitiveActivity],
    "Question",
    notes=(
        "Represents a question being asked. Question objects are unique in "
        "that they are an extension of both Content and IntransitiveActivity. "
        "That is, the Question object is an Activity but the direct object is "
        "the question itself."))

Event = ASType(
    as_uri("Event"), [Object],
    "Event",
    notes=(
        "Represents any kind of event."))

Place = ASType(
    as_uri("Place"), [Object],
    "Place",
    notes=(
        "Represents a logical or physical location. "
        "See 3.3.2 Representing Places [of ActivityStreams 2.0 Vocabulary "
        "document] for additional information."))

Mention = ASType(
    as_uri("Mention"), [Link],
    "Mention",
    notes=(
        "A specialized Link that represents an @mention."))

Profile = ASType(
    as_uri("Profile"), [Content],
    "Profile",
    notes=(
        "A Profile is a content object that describes another Object, "
        "typically used to describe Actor, objects. The describes property "
        "is used to reference the object being described by the profile."))




# Core definition Vocab and basic environment
# ===========================================

CoreVocab = ASVocab(
    [Object, Link, Activity, IntransitiveActivity, Actor, Collection,
     OrderedCollection, CollectionPage, OrderedCollectionPage,
     Accept, TentativeAccept, Add, Arrive, Create, Delete,
     Follow, Ignore, Join, Leave, Like, Offer, Invite, Reject,
     TentativeReject, Remove, Undo, Update, Experience, View,
     Listen, Read, Move, Travel, Announce, Block, Flag, Dislike,
     Application, Group, Organization, Person, Process, Service,
     Relationship, Content, Article, Album, Folder, Story, Document,
     Audio, Image, Video, Note, Page, Question, Event, Place, Mention,
     Profile])

BasicEnv = Environment(
    # @@: Maybe this one should be implied?
    vocabs=[CoreVocab],
    shortids=shortids_from_vocab(CoreVocab),
    c_accessors=shortids_from_vocab(CoreVocab))
# alias
Env = BasicEnv



## Activipy --- ActivityStreams 2.0 implementation and validator for Python
## Copyright © 2015 Christopher Allan Webber <cwebber@dustycloud.org>
##
## This file is part of Activipy, which is GPLv3+ or Apache v2, your option
## (see COPYING); since that means effectively Apache v2 here's those headers
##
## Apache v2 header:
##   Licensed under the Apache License, Version 2.0 (the "License");
##   you may not use this file except in compliance with the License.
##   You may obtain a copy of the License at
##
##       http://www.apache.org/licenses/LICENSE-2.0
##
##   Unless required by applicable law or agreed to in writing, software
##   distributed under the License is distributed on an "AS IS" BASIS,
##   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##   See the License for the specific language governing permissions and
##   limitations under the License.

import copy

import pytest

from activipy import core, vocab




# Fake inheritance tree below

def fake_type_uri(type_name):
    return "http://example.org/ns#" + type_name
    
ASObject = core.ASType(fake_type_uri("object"), [], "Object")
ASLink = core.ASType(fake_type_uri("link"), [], "Link")

ASActivity = core.ASType(fake_type_uri("activity"), [ASObject], "Activity")
ASPost = core.ASType(fake_type_uri("post"), [ASActivity], "Post")
ASDelete = core.ASType(fake_type_uri("delete"), [ASActivity], "Delete")

ASCollection = core.ASType(
    fake_type_uri("collection"), [ASObject], "Collection")
ASOrderedCollection = core.ASType(
    fake_type_uri("orderedcollection"),
    [ASCollection],
    "OrderedCollection")
ASCollectionPage = core.ASType(
    fake_type_uri("collectionpage"),
    [ASCollection],
    "CollectionPage")
ASOrderedCollectionPage = core.ASType(
    fake_type_uri("orderedcollectionpage"),
    [ASOrderedCollection, ASCollectionPage],
    "OrderedCollectionPage")

# BS testing widget
ASWidget = core.ASType(
    fake_type_uri("widget"),
    [ASObject],
    "Widget")

ASFancyWidget = core.ASType(
    fake_type_uri("fancywidget"),
    [ASWidget],
    "FancyWidget")


# Example vocab and environment
ExampleVocab = core.ASVocab(
    [ASObject, ASLink,
     ASActivity, ASPost, ASDelete,
     ASCollection, ASOrderedCollection, ASCollectionPage,
     ASOrderedCollectionPage,
     ASWidget, ASFancyWidget])

ExampleEnv = core.Environment(
    vocabs=[ExampleVocab],
    shortids=core.shortids_from_vocab(ExampleVocab),
    c_accessors=core.shortids_from_vocab(ExampleVocab))




# Basic tests

def test_inheritance_list():
    # Should just be itself
    assert core.astype_inheritance_list(ASObject) == \
        [ASObject]
    assert core.astype_inheritance_list(ASLink) == \
        [ASLink]

    # Should be itself, then its parent
    assert core.astype_inheritance_list(ASActivity) == \
        [ASActivity, ASObject]
    assert core.astype_inheritance_list(ASCollection) == \
        [ASCollection, ASObject]

    # A slightly longer inheritance chain
    assert core.astype_inheritance_list(ASPost) == \
        [ASPost, ASActivity, ASObject]
    assert core.astype_inheritance_list(ASDelete) == \
        [ASDelete, ASActivity, ASObject]
    assert core.astype_inheritance_list(ASOrderedCollection) == \
        [ASOrderedCollection, ASCollection, ASObject]
    assert core.astype_inheritance_list(ASCollectionPage) == \
        [ASCollectionPage, ASCollection, ASObject]

    # Multiple inheritance!  Egads.
    # ... this clearly demonstrates our present depth-first
    # traversal.  A breadth-first traversal would mean
    # ASCollectionPage would go before ASCollection, which may be more
    # to a user's expectations.
    assert core.astype_inheritance_list(ASOrderedCollectionPage) == \
        [ASOrderedCollectionPage, ASOrderedCollection,
         ASCollectionPage, ASCollection, ASObject,]

    # does the property version also work?
    assert ASOrderedCollectionPage.inheritance_chain == \
        [ASOrderedCollectionPage, ASOrderedCollection,
         ASCollectionPage, ASCollection, ASObject]

    # What about composite core
    assert core.astype_inheritance_list(
        ASFancyWidget, ASOrderedCollectionPage) == [
            ASFancyWidget, ASWidget,
            ASOrderedCollectionPage, ASOrderedCollection,
            ASCollectionPage, ASCollection, ASObject]
    


ROOT_BEER_NOTE_JSOBJ = {
    "@type": "Create",
    "@id": "http://tsyesika.co.uk/act/foo-id-here/",
    "actor": {
        "@type": "Person",
        "@id": "http://tsyesika.co.uk/",
        "displayName": "Jessica Tallon"},
    "to": ["acct:cwebber@identi.ca",
           "acct:justaguy@rhiaro.co.uk",
           "acct:ladyaeva@hedgehog.example"],
    "object": {
        "@type": "Note",
        "@id": "http://tsyesika.co.uk/chat/sup-yo/",
        "content": "Up for some root beer floats?"}}

ROOT_BEER_NOTE_MIXED_ASOBJ = {
    "@type": "Create",
    "@id": "http://tsyesika.co.uk/act/foo-id-here/",
    "actor": core.ASObj({
        "@type": "Person",
        "@id": "http://tsyesika.co.uk/",
        "displayName": "Jessica Tallon"}),
    "to": ["acct:cwebber@identi.ca",
           "acct:justaguy@rhiaro.co.uk",
           "acct:ladyaeva@hedgehog.example"],
    "object": core.ASObj({
        "@type": "Note",
        "@id": "http://tsyesika.co.uk/chat/sup-yo/",
        "content": "Up for some root beer floats?"})}

ROOT_BEER_NOTE_VOCAB = vocab.Create(
    "http://tsyesika.co.uk/act/foo-id-here/",
    actor=vocab.Person(
        "http://tsyesika.co.uk/",
        displayName="Jessica Tallon"),
    to=["acct:cwebber@identi.ca",
        "acct:justaguy@rhiaro.co.uk",
        "acct:ladyaeva@hedgehog.example"],
    object=vocab.Note(
        "http://tsyesika.co.uk/chat/sup-yo/",
        content="Up for some root beer floats?"))


def _looks_like_root_beer_note(jsobj):
    return (
        len(jsobj) == 5 and
        jsobj["@type"] == "Create" and
        jsobj["@id"] == "http://tsyesika.co.uk/act/foo-id-here/" and
        isinstance(jsobj["actor"], dict) and
        len(jsobj["actor"]) == 3 and
        jsobj["actor"]["@type"] == "Person" and
        jsobj["actor"]["@id"] == "http://tsyesika.co.uk/" and
        jsobj["actor"]["displayName"] == "Jessica Tallon" and
        isinstance(jsobj["to"], list) and
        jsobj["to"] == ["acct:cwebber@identi.ca",
                        "acct:justaguy@rhiaro.co.uk",
                        "acct:ladyaeva@hedgehog.example"] and
        isinstance(jsobj["object"], dict) and
        len(jsobj["object"]) == 3 and
        jsobj["object"]["@type"] == "Note" and
        jsobj["object"]["@id"] == "http://tsyesika.co.uk/chat/sup-yo/" and
        jsobj["object"]["content"] == "Up for some root beer floats?")


def test_deepcopy_jsobj_in():
    # We'll mutate later, so let's make a copy of this
    root_beer_note = copy.deepcopy(ROOT_BEER_NOTE_JSOBJ)
    assert _looks_like_root_beer_note(root_beer_note)

    # Test copying a compicated datastructure
    copied_root_beer_note = core.deepcopy_jsobj_in(
        root_beer_note, vocab.BasicEnv)
    assert _looks_like_root_beer_note(copied_root_beer_note)

    # Mutate
    root_beer_note["to"].append("sneaky@mcsneakers.example")
    assert not _looks_like_root_beer_note(root_beer_note)
    # but things are still as they were in our copy right??
    assert _looks_like_root_beer_note(copied_root_beer_note)

    # Test nested asobj copying
    assert _looks_like_root_beer_note(
        core.deepcopy_jsobj_in(
            ROOT_BEER_NOTE_MIXED_ASOBJ,
            vocab.BasicEnv))

# TODO: test_deepcopy_jsobj_out():


def test_vocab_constructor():
    assert isinstance(
        ROOT_BEER_NOTE_VOCAB,
        core.ASObj)

    assert _looks_like_root_beer_note(
        ROOT_BEER_NOTE_VOCAB.json())


ROOT_BEER_NOTE_ASOBJ = core.ASObj({
    "@type": "Create",
    "@id": "http://tsyesika.co.uk/act/foo-id-here/",
    "actor": core.ASObj({
        "@type": "Person",
        "@id": "http://tsyesika.co.uk/",
        "displayName": "Jessica Tallon"}),
    "to": ["acct:cwebber@identi.ca",
           "acct:justaguy@rhiaro.co.uk"],
    "object": core.ASObj({
        "@type": "Note",
        "@id": "http://tsyesika.co.uk/chat/sup-yo/",
        "content": "Up for some root beer floats?"})})


def test_asobj_keyaccess():
    assert ROOT_BEER_NOTE_ASOBJ["@type"] == "Create"
    assert ROOT_BEER_NOTE_ASOBJ["@id"] == \
        "http://tsyesika.co.uk/act/foo-id-here/"

    # Accessing things that look like asobjects
    # will return asobjects
    assert isinstance(
        ROOT_BEER_NOTE_ASOBJ["object"],
        core.ASObj)

    # However, we should still be able to get the
    # dictionary edition by pulling down .json()
    assert isinstance(
        ROOT_BEER_NOTE_ASOBJ.json()["object"],
        dict)

    # Traversal of traversal should work
    assert ROOT_BEER_NOTE_ASOBJ["object"]["content"] == \
        "Up for some root beer floats?"


def test_handle_one():
    # Fake value boxes
    received_args = []
    received_kwargs = []

    def test_method1(*args, **kwargs):
        received_args.append(args)
        received_kwargs.append(kwargs)
        return "Got it!"

    def test_method2(*args, **kwargs):
        # we should never hit this
        assert False
        return "skip me!"

    # well if we got this far we never hit test_method2, which is good
    # did we run test_method1 then?
    our_jsobj = ASWidget(foo="bar")
    handler = core.handle_one(
        [(test_method1, ASWidget), (test_method2, ASObject)],
        our_jsobj)
    assert handler(1, 2, foo="bar") == "Got it!"
    assert received_args[0] == (our_jsobj, 1, 2)
    assert received_kwargs[0] == {"foo": "bar"}

    # Okay, let's try running with no methods available
    with pytest.raises(core.NoMethodFound):
        core.handle_one([], our_jsobj)

    # Let's try running something with a custom fallback...
    # ... this fallback drinks the half empty glass
    glass = ["half_empty"]
    def drink_glass(asobj):
        # down the hatch
        glass.pop()
    core.handle_one([], our_jsobj, _fallback=drink_glass)
    assert len(glass) == 0  # if this passes, the pessimists win


def test_handle_map():
    def test_one(*args, **kwargs):
        return (1, args, kwargs)

    def test_two(*args, **kwargs):
        return (2, args, kwargs)

    def test_three(*args, **kwargs):
        return (3, args, kwargs)

    our_asobj = ASWidget(snorf="snizzle")
    handler = core.handle_map([(test_one, ASFancyWidget),
                                (test_two, ASWidget), (test_three, ASObject)],
                               our_asobj)
    result = handler("one", "two", "three", lets="go!")
    assert result[0][0] == 1
    assert result[1][0] == 2
    assert result[2][0] == 3
    for r in result:
        assert r[1] == (our_asobj, "one", "two", "three")
        assert r[2] == {"lets": "go!"}


def test_handle_fold():
    def test_one(asobj, val, location):
        return val + "one %s... " % location

    def test_two(asobj, val, location):
        return val + "two %s... " % location

    def test_three(asobj, val, location):
        return val + "three %s... " % location

    our_asobj = ASWidget(snorf="snizzle")
    handler = core.handle_fold([(test_one, ASFancyWidget),
                                 (test_two, ASWidget), (test_three, ASObject)],
                                our_asobj)
    result = handler("Counting down! ", "mississippi")
    assert result == (
        "Counting down! one mississippi... "
        "two mississippi... three mississippi... ")

    # Now test breaking out early
    def test_two_breaks_out(asobj, val, location):
        return core.HaltIteration(val + "two %s... " % location)

    handler = core.handle_fold([(test_one, ASFancyWidget),
                                 (test_two_breaks_out, ASWidget),
                                 (test_three, ASObject)],
                                our_asobj)
    result = handler("Counting down! ", "mississippi")
    # we never get to three in this version
    assert result == (
        "Counting down! one mississippi... "
        "two mississippi... ")




# Methods testing
# ===============

### begin testing methods ###

save = core.MethodId("save", "Save things", core.handle_one)
get_things = core.MethodId("get_things", "Build up a map of stuff",
                           core.handle_map)
# This one's name intentionally different than its variable name
# for testing reasons
combine_things = core.MethodId("combine", "combine things", core.handle_fold)

def _object_save(asobj, db):
    db[asobj["@id"]] = ("saved as object", asobj)

def _widget_save(asobj, db):
    db[asobj["@id"]] = ("saved as widget", asobj)

def _object_get_things(asobj):
    return "objects are fun"

def _activity_get_things(asobj):
    return "activities are neat"

def _post_get_things(asobj):
    return "posts are cool"

def _collection_combine_things(asobj, val, chant):
    return val + chant + ", my friend, and remember us collected\n"
    
def _ordered_collection_combine_things(asobj, val, chant):
    return val + chant + ", my dear, and cherish the order\n"

def _ordered_collection_page_combine_things(asobj, val, chant):
    return val + chant + ", my sweet, a new page is turning\n"

### end testing methods ###

MethodEnv = core.Environment(
    vocabs=[ExampleVocab],
    methods={
        (save, ASObject): _object_save,
        (save, ASWidget): _widget_save,
        (get_things, ASObject): _object_get_things,
        (get_things, ASActivity): _activity_get_things,
        (get_things, ASPost): _post_get_things,
        (combine_things, ASCollection): _collection_combine_things,
        (combine_things, ASOrderedCollection): (
            _ordered_collection_combine_things),
        (combine_things, ASOrderedCollectionPage): (
            _ordered_collection_page_combine_things)},
    shortids=core.shortids_from_vocab(ExampleVocab),
    c_accessors=core.shortids_from_vocab(ExampleVocab))


def test_environment_c_access():
    widget = MethodEnv.c.Widget(foo="bar")
    assert widget["foo"] == "bar"
    assert widget.types_astype == [ASWidget]
    assert widget.env == MethodEnv
    assert MethodEnv.c.Widget.astype == ASWidget


def test_environment_m_access_handle_one():
    db = {}
    widget = MethodEnv.c.Widget("fooid:12345")
    MethodEnv.m.save(widget, db)
    assert db["fooid:12345"] == ("saved as widget", widget)

    foo_object = MethodEnv.c.Object("fooid:00001")
    MethodEnv.m.save(foo_object, db)
    assert db["fooid:00001"] == ("saved as object", foo_object)

    collection = MethodEnv.c.Collection("fooid:8888")
    MethodEnv.m.save(collection, db)
    assert db["fooid:8888"] == ("saved as object", collection)


def test_environment_m_access_handle_map():
    result = MethodEnv.m.get_things(MethodEnv.c.Widget("fooid:12345"))
    assert result == ["objects are fun"]

    result = MethodEnv.m.get_things(MethodEnv.c.Activity("fooid:0202"))
    assert result == ["activities are neat", "objects are fun"]

    result = MethodEnv.m.get_things(MethodEnv.c.Delete("fooid:0303"))
    assert result == ["activities are neat", "objects are fun"]

    result = MethodEnv.m.get_things(MethodEnv.c.Post("fooid:0808"))
    assert result == ["posts are cool", "activities are neat",
                      "objects are fun"]


def test_environment_m_access_handle_fold():
    # This test function is kinda disturbing
    result = MethodEnv.m.combine(MethodEnv.c.Collection("fooid:1"), "",
                                 "Huzzah")
    assert result == "Huzzah, my friend, and remember us collected\n"
    
    result = MethodEnv.m.combine(MethodEnv.c.OrderedCollection("fooid:2"), "",
                                 "Huzzah")
    assert result == (
        "Huzzah, my dear, and cherish the order\n"
        "Huzzah, my friend, and remember us collected\n")

    result = MethodEnv.m.combine(MethodEnv.c.OrderedCollectionPage("fooid:2"),
                                 "", "Rejoice")
    assert result == (
        "Rejoice, my sweet, a new page is turning\n"
        "Rejoice, my dear, and cherish the order\n"
        "Rejoice, my friend, and remember us collected\n")

    # nothing for this one
    result = MethodEnv.m.combine(MethodEnv.c.Widget("fooid:12345"), "",
                                 "Huzzah")
    assert result == ""
    



## Activipy --- ActivityStreams 2.0 implementation and testing for Python
## Copyright © 2015  Christopher Allan Webber <cwebber@dustycloud.org>
##
## This file is part of Activipy, which is GPLv3+ or Apache v2, your option
## (see COPYING); since that means effectively Apache v2...
##
## Apache v2 header:
##   Licensed under the Apache License, Version 2.0 (the "License");
##   you may not use this file except in compliance with the License.
##   You may obtain a copy of the License at
##
##       http://www.apache.org/licenses/LICENSE-2.0
##
##   Unless required by applicable law or agreed to in writing, software
##   distributed under the License is distributed on an "AS IS" BASIS,
##   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##   See the License for the specific language governing permissions and
##   limitations under the License.

from setuptools import setup, find_packages

setup(
    name="activipy",
    version="0.1",
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
    zip_safe=False,
    include_package_data=True,
    install_requires=[
        "PyLD",
        "pytest",
        "sphinx",
        ],
    # @@: Can we reproduce this in Guix?
    entry_points="""\
        [console_scripts]
        activipy_tester = activipy.testcli:main
        """,
    license="Apache v2",
    author="Christopher Allan Webber",
    author_email="cwebber@dustycloud.org",
    description="ActivityStreams 2.0 implementation and testing for Python",
    long_description="""An ActivityStreams 2.0 implementation for Python.
Provides an easy API for building ActivityStreams 2.0 based applications
as well as a test suite for testing ActivityStreams 2.0 libraries against.""",
    classifiers=[
        # @@: Might need to drop v2
        "License :: OSI Approved :: Apache Software License"])



## Activipy --- ActivityStreams 2.0 implementation and validator for Python
## Copyright © 2015 Christopher Allan Webber <cwebber@dustycloud.org>
##
## This file is part of Activipy, which is GPLv3+ or Apache v2, your option
## (see COPYING); since that means effectively Apache v2 here's those headers
##
## Apache v2 header:
##   Licensed under the Apache License, Version 2.0 (the "License");
##   you may not use this file except in compliance with the License.
##   You may obtain a copy of the License at
##
##       http://www.apache.org/licenses/LICENSE-2.0
##
##   Unless required by applicable law or agreed to in writing, software
##   distributed under the License is distributed on an "AS IS" BASIS,
##   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##   See the License for the specific language governing permissions and
##   limitations under the License.


import json
import sys
import argparse
from collections import namedtuple, OrderedDict

from .core import Activity, InvalidActivity

class UserError(Exception): pass
class InvalidInput(UserError): pass




# Dump command
# ============

def dump_cli(args):
    try:
        asobj = json.loads(args.asobj)
    except ValueError:
        raise InvalidInput(
            "Not valid json: %s" % args.asobj)

    activity = Activity.from_json(asobj)
    try:
        activity.validate()
    except InvalidActivity as error:
        raise InvalidInput(str(error))


def dump_setup_subparser(subparser):
    subparser.add_argument(
        "asobj",
        help="ActivityStreams object, as json")




# Build command
# ==============

def build_cli():
    pass


def build_setup_subparser(subparser):
    pass




# Testdriver command
# ==================

def testdriver_cli():
    pass


def testdriver_setup_subparser(subparser):
    pass




# Build CLI
# =========

Command = namedtuple("command", ["cli_proc", "setup_subparser"])

SUBCOMMANDS_MAP = OrderedDict([
    ("dump", Command(
        dump_cli, dump_setup_subparser)),
    ("build", Command(
        build_cli, build_setup_subparser)),
    ("testdriver", Command(
        testdriver_cli, testdriver_setup_subparser))])


def main():
    parser = argparse.ArgumentParser(
        # @@: this sucks as a description
        description="Test for activitystreams correctness")

    subparsers = parser.add_subparsers(dest="subparser_name")

    for subcommand_key, subcommand_cmd in SUBCOMMANDS_MAP.items():
        subcmd_parser = subparsers.add_parser(subcommand_key)
        subcommand_cmd.setup_subparser(subcmd_parser)

    args = parser.parse_args()
    if not args.subparser_name:
        parser.print_help()
        sys.exit(1)

    try:
        subcmd_proc = SUBCOMMANDS_MAP[args.subparser_name].cli_proc
        subcmd_proc(args)
    except UserError as error:
        print(error)
        # it's only half evil that the user made this mistake,
        # is my guess
        sys.exit(333)


if __name__ == "__main__":
    main()

